// File Portpin.h for CC1B
#ifndef _PORTPIN_H_
#define _PORTPIN_H_
//include other libraries here
#pragma library 1

// port functions

/**
 * Update DDR register
 * This function writes the DDR copy register for the specified port to the appropiate chip DDR register.
 * It is used by the pin and port direction functions.
 * This function prevents possible glitches on port pins (in case the interrupt routine also accesses the DDR registers),
 * by inserting a few NOPs when a RTCC rollover interrupt would occur between reading the DDR copy register
 * and updating the chip DDR register.
 *
 * @param W Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 */
void DDRupdate(char W) {
  W &= 0x0F;
  switch (W) { //update correct DDR register without possible glitches due to interrupt
    case PORT_RA:
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
                  W = 0xF8;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
                  W = 0xF7;
#endif
                  W = RTCC - W; //if RTCC != W at start of this instruction, then 4 clock ticks
	              if (Zero_) { //else 5 clock ticks, so interrupt occurs before MOV W,DDRACOPY
	                nop();
	                nop();
	                nop();
	              }
	              DDRA = DDRACOPY;
                  break;
    case PORT_RB:
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
                  W = 0xF8;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
                  W = 0xF7;
#endif
                  W = RTCC - W;
	              if (Zero_) {
	                nop();
	                nop();
	                nop();
	              }
	              DDRB = DDRBCOPY;
                  break;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RC:
#if defined _CHIP_SX28_
                  W = 0xF8;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
                  W = 0xF7;
#endif
                  W = RTCC - W;
	              if (Zero_) {
	                nop();
	                nop();
	                nop();
	              }
	              DDRC = DDRCCOPY;
                  break;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RD:
                  W = 0xF7;
                  W = RTCC - W;
	              if (Zero_) {
	                nop();
	                nop();
	                nop();
	              }
	              DDRD = DDRDCOPY;
                  break;
    case PORT_RE:
                  W = 0xF7;
                  W = RTCC - W;
	              if (Zero_) {
	                nop();
	                nop();
	                nop();
	              }
	              DDRE = DDRECOPY;
                  break;
#endif
  }
}

/**
 * Get port direction
 *
 * @param port Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 * @return direction Bits specify direction for corresponding pins
 *                                 0 = output, 1 = input
 */
char getPortDirection(char port) {
  W = port | 0xF0;
#pragma update_FSR 0
  FSR = W;
  W = INDF; //get DDRACOPY to DDRECOPY
#pragma update_FSR 1
  return W;
}

/**
 * Set port direction
 *
 * @param port Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 * @param direction Bits specify direction for corresponding pins
 *                                 0 = output, 1 = input
 */
void setPortDirection(char port, char direction) {
  _MainTemp = direction;
  W = port | 0xF0;
#pragma update_FSR 0
  FSR = W;
  W = _MainTemp;
  INDF = W; //update DDRACOPY to DDRECOPY
#pragma update_FSR 1
  DDRupdate(port);
}

/**
 * Set port level
 *
 * @param port Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 * @param level Bits specify cmos (2.5V) or ttl (1.4V) level for corresponding pins
 *                         0 = cmos, 1 = ttl
 */
void setPortLevel(char port, char level) {
  switch (port & 0xF0) {
    case PORT_RA: LVL_A = level;
	              break;
    case PORT_RB: LVL_B = level;
	              break;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RC: LVL_C = level;
	              break;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RD: LVL_D = level;
	              break;
    case PORT_RE: LVL_E = level;
	              break;
#endif
  }
}

/**
 * Set port pullup
 *
 * @param port Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 * @param pullup Bits specify internal pullup resistors setting for corresponding pins
 *                           0 = enabled, 1 = disabled
 */
void setPortPullup(char port, char pullup) {
  switch (port & 0xF0) {
    case PORT_RA: PLP_A = pullup;
	              break;
    case PORT_RB: PLP_B = pullup;
	              break;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RC: PLP_C = pullup;
	              break;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RD: PLP_D = pullup;
	              break;
    case PORT_RE: PLP_E = pullup;
	              break;
#endif
  }
}

/**
 * Set port schmitt trigger
 *
 * @param port Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 * @param trigger Bits specify schmitt trigger inputs setting for corresponding pins
 *                             0 = enabled, 1 = disabled
 */
void setPortSchmittTrigger(char port, char trigger) {
  switch (port & 0xF0) {
    case PORT_RB: ST_B = trigger;
	              break;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RC: ST_C = trigger;
	              break;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
    case PORT_RD: ST_D = trigger;
	              break;
    case PORT_RE: ST_E = trigger;
	              break;
#endif
  }
}

/**
 * Write value to port latch register.
 * The port direction register is not changed.
 *
 * @param port Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 * @param value Bits specify output level for corresponding pins
 *                          0 = low, 1 = high
 */
void writePort(char port, char value) {
  _MainTemp = value;
  W = port & 0x0F;
#pragma update_FSR 0
  FSR = W;
  W = _MainTemp;
  INDF = W; //write value
#pragma update_FSR 1
}

/**
 * Read value from port pins.
 *
 * @param port Port identifier PORT_RA to PORT_RE
 *                        It is also possible to use the highbyte of a portpin value, eg. PORTPIN_RA>>8
 * @return Bits specify input level at corresponding pins
 *                0 = low, 1 = high
 */
char readPort(char port) {
  W = port & 0x0F;
#pragma update_FSR 0
  FSR = W;
  W = INDF; //read value
#pragma update_FSR 1
  return W;
}

// pin functions

/**
 * Make specified port pins output.
 * This function clears specified bits in DDRACOPY to DDRECOPY and then updates the appropiate DDR register.
 * The port latch registers are not changed.
 *
 * @param portpin Specified pin or pins of the same port
 *                             A single pin can be specified, eg. PORTPIN_RA0, or an ORed combination
 *                             eg. PORTPIN_RB0 | PORTPIN_RB2 | PORTPIN_RB7 or PORTPIN_RB for all pins
 */
void setPinOutput(long portpin) {
  _MainTemp = ~portpin.low8;
  W = portpin.high8 | 0xF0;
#pragma update_FSR 0
  FSR = W;
  W = _MainTemp;
  INDF &= W; //clear selected bits of DDRACOPY to DDRECOPY
#pragma update_FSR 1
  DDRupdate(portpin.high8);
}

/**
 * Make specified port pins input.
 * This function sets specified bits in DDRACOPY to DDRECOPY and then updates the appropiate DDR register.
 * The port latch registers are not changed.
 *
 * @param portpin Specified pin or pins of the same port
 *                             A single pin can be specified, eg. PORTPIN_RA0, or an ORed combination
 *                             eg. PORTPIN_RB0 | PORTPIN_RB2 | PORTPIN_RB7 or PORTPIN_RB for all pins
 */
void setPinInput(long portpin) {
  _MainTemp = portpin.low8;
  W = portpin.high8 | 0xF0;
#pragma update_FSR 0
  FSR = W;
  W = _MainTemp;
  INDF |= W; //set selected bits of DDRACOPY to DDRECOPY
#pragma update_FSR 1
  DDRupdate(portpin.high8);
}

/**
 * Get pin direction
 *
 * @param portpin Specified pin
 *                             A single pin can be specified, eg. PORTPIN_RA0
 * @return direction Specifies direction for pin
 *                                 0 = output, 1 = input
 */
bit getPinDirection(long portpin) {
  char temp = getPortDirection(portpin.high8);
  temp &= portpin.low8;
  return !Zero_;
}

/**
 * Make specified port pins input or output.
 * This function sets or clears specified bits in DDRACOPY to DDRECOPY and then updates the appropiate DDR register.
 * The port latch registers are not changed.
 *
 * @param portpin Specified pin or pins of the same port
 *                             A single pin can be specified, eg. PORTPIN_RA0, or an ORed combination
 *                             eg. PORTPIN_RB0 | PORTPIN_RB2 | PORTPIN_RB7 or PORTPIN_RB for all pins
 * @param W 1 for inputs, 0 for outputs
 */
void setPinDirection(long portpin, char W) {
  if (W) setPinInput(portpin);
  else setPinOutput(portpin);
}

/**
 * Invert specified port pin direction.
 * This function inverts the specified bit of the port latch.
 * The port direction registers are not changed.
 *
 * @param portpin Specified pin
 *                             A single pin can be specified, eg. PORTPIN_RA0
 */
void invertPinDirection(long portpin) {
  bit value = !getPinDirection(portpin.high8);
  setPinDirection(portpin,value);
}

/**
 * Make specified port pins low.
 * This function clears specified bits in RA to RE latch register.
 * The port direction registers are not changed.
 *
 * @param portpin Specified pin or pins of the same port
 *                             A single pin can be specified, eg. PORTPIN_RA0, or an ORed combination
 *                             eg. PORTPIN_RB0 | PORTPIN_RB2 | PORTPIN_RB7 or PORTPIN_RB for all pins
 */
void setPinLow(long portpin) {
  _MainTemp = ~portpin.low8;
  W = portpin.high8 & 0x0F;
#pragma update_FSR 0
  FSR = W;
  W = _MainTemp;
  INDF &= W; //clear selected bits of RA to RE
#pragma update_FSR 1
}

/**
 * Make specified port pins high.
 * This function sets specified bits in RA to RE latch register.
 * The port direction registers are not changed.
 *
 * @param portpin Specified pin or pins of the same port
 *                             A single pin can be specified, eg. PORTPIN_RA0, or an ORed combination
 *                             eg. PORTPIN_RB0 | PORTPIN_RB2 | PORTPIN_RB7 or PORTPIN_RB for all pins
 */
void setPinHigh(long portpin) {
  _MainTemp = portpin.low8;
  W = portpin.high8 & 0x0F;
#pragma update_FSR 0
  FSR = W;
  W = _MainTemp;
  INDF |= W; //set selected bits of RA to RE
#pragma update_FSR 1
}

/**
 * Make specified port pins low or high.
 * This function clears or sets specified bits in RA to RE latch register.
 * The port direction registers are not changed.
 *
 * @param portpin Specified pin or pins of the same port
 *                             A single pin can be specified, eg. PORTPIN_RA0, or an ORed combination
 *                             eg. PORTPIN_RB0 | PORTPIN_RB2 | PORTPIN_RB7 or PORTPIN_RB for all pins
 * @param W 0 for low, 1 for high
 */
void writePin(long portpin, char W) {
  if (W) setPinHigh(portpin);
  else setPinLow(portpin);
}

/**
 * Read specified port pin.
 * This function reads the specified port input pin.
 *
 * @param portpin Specified pin
 *                             A single pin can be specified, eg. PORTPIN_RA0
 * @return level at pin input 0 for low, 1 for high
 */
bit readPin(long portpin) {
  char temp = readPort(portpin.high8);
  temp &= portpin.low8;
  return !Zero_;
}

/**
 * Toggle specified port pin.
 * This function inverts the specified bit of the port latch.
 * The port direction registers are not changed.
 *
 * @param portpin Specified pin
 *                             A single pin can be specified, eg. PORTPIN_RA0
 */
void togglePin(long portpin) {
  bit value = !readPin(portpin.high8);
  writePin(portpin,value);
}

#pragma library 0
#endif
// End of file Portpin.h
