// File objQueue.h for CC1B
#ifndef _OBJQUEUE_H_
#define _OBJQUEUE_H_
//include other libraries here
#include <memory.h>
#pragma library 1

/**
 * This library provides a queue object that can store up to 14 bytes. The queue resides in a single rambank.
 * If the queue cannot reside in a single rambank for the given address and length, an error is throwed.
 *
 * A queue is declared by using the objQueue macro
 * objQueue(Q1,0x72,10)
 * This declares queue Q1 that is located at ram address 0x72 and has a total length of 12 bytes (10 bytes data + 2 bytes overhead)
 *
 * Before using the queue it must be initialized. This is done by function objInit.
 * Note that objInit takes as parameter the queue name with _init suffix.
 * So for the above declaration use objInit(Q1_init)
 *
 * Values are stored into the queue using function objQueue_enqueue.
 * This function returns true when the value is stored, or false if the queue is full
 * if (!objQueue_enqueue(Q1,value) {
 *   //queue full
 * }
 *
 * Values are retrieved from the queue using function objQueue_dequeue.
 * This function returns true when the value is retrieved, or false if the queue is empty
 * The retrieved value is returned in W
 * if (objQueue_dequeue(Q1) {
 *   value = W;
 * }
 *
 * The queue can be cleared by using the function objQueue_clear
 */

//queue buffer layout
//
// +----+----+
// |size| num|  size = number of databytes in queue buffer, num = number of values in queue 
// +----+----+
// |head|tail|  head = position last stored value, tail = position before value to retrieve
// +----+----+
// | d a t a |
// | d a t a |

//#if (len < 3) || (sizeof(name) > len) \
//#error Invalid queue length or queue does not fit in variable \
//#endif \

#define objQueue(name,len) \
#pragma cdata. ##name##_init = 2,(len-2)<<4,0

/** clear queue
 * @param W Queue address
 */
void objQueue_clear(char W) {
#pragma update_FSR 0
  FSR = W;
  W = 0xF0;
  INDF &= W;  //clear num
  FSR++;
  INDF = 0; //clear headtail
#pragma update_FSR 1
}

/** enqueue value
 * @param addr Queue address
 * @param W Value to store
 * @return True if value stored, false if queue full
 */
bit objQueue_enqueue(char addr, char W) {
  _MainTemp = W;
  W = addr;
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF);
  W ^= INDF;
  if (!Zero_) {
    INDF++; //inc num
    FSR++;
    INDF += 0x10; //inc head
    FSR--;
    W = INDF; //get size
    FSR++;
    W ^= INDF; //xor size with head
    W &= 0xF0;
    W = 0x0F;
    if (Zero_) INDF &= W; //if head == size then clear head
    W = swap(INDF); //get head in lownib
    W &= 0x0F;
    FSR++; //point to queue data begin
    FSR += W; //head position
    INDF = _MainTemp; //store value
	return 1;
  }
#pragma update_FSR 1
  return 0;
}

/** dequeue value
 * @param W Queue address
 * @return True if value retrieved (W holds value), false if queue empty
 */
bit objQueue_dequeue(char W) {
#pragma update_FSR 0
  FSR = W;
  W = INDF; //load size+num
  W &= 0x0F; //test num
  if (!Zero_) {
    INDF--; //dec num
	W = swap(INDF); //get size in lownib
	FSR++;
	INDF++; //inc tail
	W ^= INDF; //xor size with tail
	W &= 0x0F;
	W = 0xF0;
	if (Zero_) INDF &= W; //if tail == size then clear tail
	W = INDF; //get tail in lownib
	W &= 0x0F;
	FSR++; //point to queue data begin
	FSR += W; //tail position
	W = INDF; //get value
	return 1;
  }
#pragma update_FSR 1
  return 0;
}

#pragma library 0
#endif
// End of file objQueue.h
